﻿// License: The MIT License (MIT) Copyright (c) 2011 Barend Gehrels

// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using System.Windows.Threading;

namespace GrooveSharp
{

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private Ownskit.Utils.KeyboardListener m_KeyboardListener = null;
        private DispatcherTimer m_KeyboardTimer;
        private Key m_KeyboardListenedKey;

        private DispatcherTimer m_ActivityTimer;
        private int m_ActivityCounter = 0;

        // Currently this "MainWindow" keeps control over the current playing list, that might change.
        private int m_TrackIndex = -1;
        private List<QueuedTrack> m_TrackQueue = new List<QueuedTrack>();

        private SizeManager m_SizeManager = new SizeManager();
        private IList<FolderWatcher> m_FolderWatchers = new List<FolderWatcher>();
        private IList<XmlWatcher> m_XmlWatchers = new List<XmlWatcher>();

        private Announcer m_announcer = new Announcer();
        private UpdatePlayDates m_updatePlayDates = new UpdatePlayDates();

        private IList<IPlayer> m_players = new List<IPlayer>();

        public MainWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            foreach (var path in Db.Connection().Select<GroovePath>("select * from paths where is_active=1"))
            {
                m_FolderWatchers.Add(new FolderWatcher(path));
            }
            m_XmlWatchers.Add(new XmlWatcher(@"c:\users\Barend\Documents\iPad.xml"));

            m_ActivityTimer = new DispatcherTimer();
            m_ActivityTimer.Interval = TimeSpan.FromSeconds(1.0);
            m_ActivityTimer.Tick += new EventHandler(activityTimer_Tick);
            m_ActivityTimer.Start();

            m_KeyboardTimer = new DispatcherTimer();
            m_KeyboardTimer.Interval = TimeSpan.FromSeconds(0.1);
            m_KeyboardTimer.Tick += new EventHandler(keyboardTimer_Tick);

            imageMoveGrip.Visibility = Visibility.Hidden;

#if DEBUG
            buttonTest.Visibility = Visibility.Visible;
#endif

            theCollection.Scan();

            // Uncommenting means listening, which triggers Kaspersky.
            // http://www.ehow.com/facts_7623515_pdm-keylogger.html
            // http://support.kaspersky.com/kav2011/settings/proactive?print=true&qid=208284446
            // Here is how to turn it off:
            // http://utils.kaspersky.com/special/kis2012/29_kis2012_proactive_defence_en.pdf
            // But it could also be signed.


            if (SettingManager.UseKeyboardHook())
            {
                m_KeyboardListener = new Ownskit.Utils.KeyboardListener();
            }
            if (m_KeyboardListener != null)
            {
                m_KeyboardListener.KeyDown += new Ownskit.Utils.RawKeyEventHandler(KeyboardListener_KeyDown);
            }

            m_players.Add(theNormalInfo);
            m_players.Add(theCompactInfo);
            m_players.Add(theCurrentPlayList);
            m_players.Add(m_announcer);
            m_players.Add(m_updatePlayDates);
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            if (m_KeyboardListener != null)
            {
                m_KeyboardListener.Dispose();
                m_KeyboardListener = null;
            }

            // Stop which is necessary, at least play datum to store end-date
            m_updatePlayDates.Stop();
        }


        private void Play(QueuedTrack track)
        {
            try
            {
                // First check to update later.
                bool wasPlaying = theMainPlayer.Playing();

                // The main player plays as fast as possible - to avoid gaps between music. It is audible.
                // It does not need info/bitmaps so it plays before.
                TrackInfoHelper.TrackInfo info = null;
                theMainPlayer.Play(track, info);

                // Now update stops - where necessary
                if (wasPlaying)
                {
                    m_updatePlayDates.Stop();
                }


                info = TrackInfoHelper.GetTrackInfo(track.Id, track.FileName);

                foreach (IPlayer player in m_players)
                {
                    player.Play(track, info);
                }

                UpdatePauseButton();
            }
            catch(Exception e)
            {
                App.ReportException(e);
            }
        }

        private void Play(int index)
        {
            // Defensive check on ID which is checked before
            if (index >= 0 && index < m_TrackQueue.Count)
            {
                Play(m_TrackQueue[index]);
            }
        }

        private void UpdatePauseButton()
        {
            BitmapImage image = new BitmapImage();
            image.BeginInit();
            image.UriSource = new Uri(String.Format("pack://application:,,/Images/{0}.gif", theMainPlayer.IsPaused() ? "play" : "pause"));
            image.EndInit();
            imagePlayPause.Source = image;
        }

        private void TogglePause()
        {
            theMainPlayer.TogglePause();
            UpdatePauseButton();
        }


        private void Next()
        {
            if (m_TrackIndex + 1 < m_TrackQueue.Count)
            {
                Play(++m_TrackIndex);
            }
        }

        private void Previous()
        {
            if (m_TrackIndex > 0)
            {
                Play(--m_TrackIndex);
            }
        }

        private void NextIfNotStarted()
        {
            if (! theMainPlayer.Playing())
            {
                Next();
            }
        }


        private void AddTrack(QueuedTrack track)
        {
            if (track.Id < 0)
            {
                int newId = Db.TrackIdFromFileName(track.FileName);
                if (newId > 0)
                {
                    track.Id = newId;
                }
            }
            
            if (track.Id > 0)
            {
                // Avoid duplicates, so return early
                foreach(var p in (from p in m_TrackQueue where p.Id == track.Id select p))
                {
                    return;
                }
            }
            m_TrackQueue.Add(track);

            theCurrentPlayList.Add(track);
        }
        private void AddTrack(string fileName)
        {
            int id = Db.TrackIdFromFileName(fileName);
            if (id == 0)
            {
                // Give track/file an ID of -1,-2,-3,-4 etc
                var fileIds = from p in m_TrackQueue where p.Id < 0 select p.Id;
                id = (fileIds.Count() > 0 ? fileIds.Min() : 0) - 1;
            }
            AddTrack(new QueuedTrack() { Id = id, FileName = fileName });
        }

        private void MusicCollection_AddTrack(object sender, RoutedEventArgs e)
        {
            foreach (QueuedTrack track in theCollection.SelectedTracks())
            {
                AddTrack(track);
            }
            NextIfNotStarted();
        }

        private void buttonNext_Click(object sender, RoutedEventArgs e)
        {
            Next();
        }


        private void mainTabContro1_SelectionChanged(object sender, System.Windows.Controls.SelectionChangedEventArgs e)
        {
            if (mainTabControl.SelectedIndex == 1)
            {
                theCollection.LoadCollectionIfEmpty();
            }
        }

        private void buttonPrev_Click(object sender, RoutedEventArgs e)
        {
            Previous();
        }

        private void buttonPause_Click(object sender, RoutedEventArgs e)
        {
            TogglePause();
        }

        private void KeyAction(Key key)
        {
            switch (key)
            {
                case System.Windows.Input.Key.BrowserHome: mainTabControl.SelectedIndex = 0; break;
                case System.Windows.Input.Key.BrowserSearch: mainTabControl.SelectedIndex = 1; break;
                case System.Windows.Input.Key.MediaPlayPause: TogglePause(); break;
                case System.Windows.Input.Key.BrowserBack: Previous(); break;
                case System.Windows.Input.Key.BrowserForward: Next(); break;
            }
        }

        private bool KeyHasAction(Key key)
        {
            switch (key)
            {
                case System.Windows.Input.Key.BrowserHome: return true;
                case System.Windows.Input.Key.BrowserSearch: return true;
                case System.Windows.Input.Key.MediaPlayPause: return true;
                case System.Windows.Input.Key.BrowserBack: return true;
                case System.Windows.Input.Key.BrowserForward: return true;
            }
            return false;
        }

        private void keyboardTimer_Tick(object sender, EventArgs e)
        {
            m_KeyboardTimer.Stop();
            KeyAction(m_KeyboardListenedKey);
        }

        void KeyboardListener_KeyDown(object sender, Ownskit.Utils.RawKeyEventArgs args)
        {
            if (KeyHasAction(args.Key))
            {
                // We cannot take action on the key immediately, because it is in another thread.
                // This is solved using a specific timer (m_KeyboardTimer)
                m_KeyboardListenedKey = args.Key;
                m_KeyboardTimer.Start();
            }
        }

        private void Window_PreviewKeyDown(object sender, System.Windows.Input.KeyEventArgs e)
        {
            if (KeyHasAction(e.Key) && m_KeyboardListener == null)
            {
                KeyAction(e.Key);
            }
        }

        private void theMainPlayer_MusicStart(object sender, RoutedEventArgs e)
        {
            UpdatePauseButton();
        }

        private void theMainPlayer_MusicFade(object sender, RoutedEventArgs e)
        {
            theNormalInfo.FadeOut();
        }

        private void theMainPlayer_MusicEnd(object sender, RoutedEventArgs e)
        {
            foreach (IPlayer player in m_players)
            {
                player.Stop();
            }
            Next();
        }

        private void buttonExit_Click(object sender, RoutedEventArgs e)
        {
            Application.Current.Shutdown();
        }

        // Users can drop MP3 files on any part of the window as well
        private void Window_Drop(object sender, DragEventArgs e)
        {
            if (e != null && e.Data != null)
            {
                object filenames = e.Data.GetData(DataFormats.FileDrop, true);
                if (filenames is IEnumerable<string>)
                {
                    foreach (string f in filenames as IEnumerable<string>)
                    {
                        if (f.ToUpper().EndsWith(".MP3"))
                        {
                            AddTrack(f);
                        }
                    }
                    e.Handled = true;
                    NextIfNotStarted();
                }
            }
        }



        // Rating
        private void SaveAndSynchronizeRating(int ratingValue)
        {
            if (m_TrackIndex >= 0 && m_TrackIndex < m_TrackQueue.Count)
            {
                int trackId = m_TrackQueue[m_TrackIndex].Id;
                if (trackId > 0)
                {
                    TrackRating tr = new TrackRating() { TrackId = trackId, Value = ratingValue };
                    Db.Connection().Insert(tr);

                    // Synchronize it in all dependant lists/infos
                    theCompactInfo.SetRating(tr);
                    theNormalInfo.SetRating(tr);
                    theCurrentPlayList.SetRating(tr);
                }
            }
        }


        private void theCompactInfo_RatingChanged(object sender, RoutedEventArgs e)
        {
            SaveAndSynchronizeRating(theCompactInfo.theRating.RatingValue);
        }

        private void theNormalInfo_RatingChanged(object sender, RoutedEventArgs e)
        {
            SaveAndSynchronizeRating(theNormalInfo.theRating.RatingValue);
        }

        private void theCurrentPlayList_RatingChanged(object sender, RoutedEventArgs e)
        {
            if (m_TrackIndex >= 0 && m_TrackIndex < m_TrackQueue.Count)
            {
                int trackId = m_TrackQueue[m_TrackIndex].Id;
                if (trackId > 0)
                {
                    if (theCurrentPlayList.LastChangedRating.TrackId == trackId)
                    {
                        SaveAndSynchronizeRating(theCurrentPlayList.LastChangedRating.Value);
                    }
                    else
                    {
                        Db.Connection().Insert(theCurrentPlayList.LastChangedRating);
                    }
                }
            }
        }


        private void theCurrentPlayList_PlayOtherTrack(object sender, RoutedEventArgs e)
        {
            int index = 0;
            foreach (QueuedTrack track in m_TrackQueue)
            {
                if (track.Id == theCurrentPlayList.SelectedId)
                {
                    m_TrackIndex = index;
                    Play(index);
                    return;
                }
                index++;
            }
        }

        private void theCurrentPlayList_OnTrackOrderChanged(object sender, RoutedEventArgs e)
        {
            var ids = theCurrentPlayList.OrderedIds;
            var newList = new List<QueuedTrack>();
            foreach (int id in theCurrentPlayList.OrderedIds)
            {
                newList.AddRange(from p in m_TrackQueue where p.Id == id select p);
            }
            if (m_TrackIndex >= 0 && m_TrackIndex < m_TrackQueue.Count)
            {
                int oldId = m_TrackQueue[m_TrackIndex].Id;

                // Update the current playing index
                int index = 0;
                foreach (var p in newList)
                {
                    if (p.Id == oldId)
                    {
                        m_TrackIndex = index;
                        break;
                    }
                    index++;
                }
            }

            m_TrackQueue = newList;
        }


        #region HIDING BUTTONS IF NO INTERACTION
        private void setVisibilityOfMany(Visibility setting)
        {
            topStack.Visibility = setting;
            tabItemPlayer.Visibility = setting;
            tabItemCollection.Visibility = setting;
            tabItemCurrentPlayList.Visibility = setting;
            tabItemSettings.Visibility = setting;
            tabItemHelp.Visibility = setting;

            // Setting bottomStack triggers a mouse-change event (curiously enough)
            //bottomStack.Visibility = setting;
            buttonPrev.Visibility = setting;
            buttonNext.Visibility = setting;
            buttonPause.Visibility = setting;

            if (this.WindowState != WindowState.Maximized)
            {
                imageMoveGrip.Visibility = setting;
            }

#if DEBUG
            buttonTest.Visibility = setting;
#endif
        }

        private void activityTimer_Tick(object sender, EventArgs e)
        {
            if (mainTabControl.SelectedIndex == 0 || m_SizeManager.IsCompact)
            {
                m_ActivityCounter++;
            }

            // At 5 seconds:
            if (m_ActivityCounter == 5)
            {
                // Collapsed does not work because triggers a mouse-change event (curiously enough)
                setVisibilityOfMany(Visibility.Hidden);
                App.Current.MainWindow.Cursor = Cursors.None;

                // DON'T uncomment this: it restarts the MP3. Alas, probably bug in WPF.
                /*if (isCompact)
                {
                    this.ResizeMode = ResizeMode.NoResize;
                }*/

            }
        }

        private void AnyInteraction()
        {
            // Reset the timer and make buttons visible
            m_ActivityCounter = 0;
            setVisibilityOfMany(Visibility.Visible);
            App.Current.MainWindow.Cursor = Cursors.Arrow;
            this.ResizeMode = ResizeMode.CanResizeWithGrip;
        }

        private void Window_PreviewTextInput(object sender, TextCompositionEventArgs e)
        {
            // This event fires on a normal key, but not on shift/ctrl/etc
            // which is handled in PreviewKeyDown
            AnyInteraction();
        }

        private void Window_PreviewMouseMove(object sender, System.Windows.Input.MouseEventArgs e)
        {
            AnyInteraction();
        }

        #endregion


        #region RESIZING
        private void buttonState_Click(object sender, RoutedEventArgs e)
        {
            bool setMaximized = this.WindowState != WindowState.Maximized;
            this.WindowState = setMaximized ? WindowState.Maximized : WindowState.Normal;
            //imageMoveGrip.Visibility = setMaximized ? Visibility.Hidden : Visibility.Visible;
            //this.Topmost = !setMaximized;
        }

        private void SetButtonForCompact(Button b, Image i, bool compact)
        {
            b.Width = compact ? 16 : 40;
            b.Height = compact ? 16 : 40;
            i.Width = compact ? 8 : 18;
            i.Height = compact ? 8 : 18;

            b.Margin = new Thickness(compact ? 3 : 10);
        }

        private void SetCompactness(bool toCompact)
        {
            if (toCompact)
            {
                mainTabControl.Visibility = Visibility.Collapsed;
                theCompactInfo.Visibility = Visibility.Visible;
                theMainPlayer.mediaProgressBar.Height = 5;
                panelHeader.Margin = new Thickness(0, 2, 2, 0);
            }
            else
            {
                mainTabControl.Visibility = Visibility.Visible;
                theCompactInfo.Visibility = Visibility.Collapsed;
                theMainPlayer.mediaProgressBar.Height = 15;
                panelHeader.Margin = new Thickness(0, 8, 8, 0);

            }
            SetButtonForCompact(buttonPrev, imagePrev, toCompact);
            SetButtonForCompact(buttonNext, imageNext, toCompact);
            SetButtonForCompact(buttonPause, imagePlayPause, toCompact);
            SetButtonForCompact(buttonTest, imagePlayPause, toCompact);
        }

        private void buttonCompact_Click(object sender, RoutedEventArgs e)
        {
            m_SizeManager.ToggleCompactness(this);
            SetCompactness(m_SizeManager.IsCompact);
        }

        private void Window_StateChanged(object sender, EventArgs e)
        {
            System.Diagnostics.Debug.Print("State changed");
            bool nowMaximized = this.WindowState == WindowState.Maximized;
            imageMoveGrip.Visibility = nowMaximized ? Visibility.Hidden : Visibility.Visible;
            this.Topmost = !nowMaximized;

            SetCompactness(m_SizeManager.IsCompact && !nowMaximized);
        }

        private void Window_LocationChanged(object sender, EventArgs e)
        {
            m_SizeManager.SizeOrLocationChanged(this);
        }

        private void Window_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            m_SizeManager.SizeOrLocationChanged(this);
        }

        #endregion


        #region DRAG MOVE USING GRIP

        // http://social.msdn.microsoft.com/Forums/en/wpf/thread/f9ff4c24-506b-4683-b023-34a73f48feee
        private bool m_Moving = false;

        private void moveGripImage_MouseLeftButtonDown(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            m_Moving = true;
        }

        private void moveGripImage_MouseLeftButtonUp(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            // We set m_moving to false but this event is indeed not fired at all:
            // http://stackoverflow.com/questions/1917176/c-sharp-wpf-dragmove-and-click
            // Anyway, it is working correctly because of MouseLeave
            m_Moving = false;
        }

        private void moveGripImage_MouseLeave(object sender, MouseEventArgs e)
        {
            m_Moving = false;
        }

        private void moveGripImage_MouseMove(object sender, MouseEventArgs e)
        {
            if (m_Moving)
            {
                this.DragMove();
            }

        }

        #endregion


        #region TEST CODE ONLY
        // ------------------------------------------------------------------------------------
        private void AddFolderToQueue(string folderName)
        {
            foreach (string file in Directory.GetFiles(folderName, "*.mp3"))
            {
                AddTrack(new QueuedTrack() { FileName = file });
            }
        }

        private void buttonTest_Click(object sender, RoutedEventArgs e)
        {
            // Short tracks
            if (m_TrackQueue.Count == 0)
            {
                AddTrack(@"c:\users\Barend\Music\Prince\Gold Experience\06 NPG Operator.mp3");
                AddTrack(@"c:\users\Barend\Music\David Bowie\Outside [Australia]\17 Segue- Nathan Adler.mp3");
                AddTrack(@"c:\users\Barend\Music\Pink Floyd\The Wall Disc 2\11 Stop.mp3");

                // Albums
                //addFolderToQueue(@"c:\Users\Barend\Music\Radiohead\OK Computer");
                //addFolderToQueue(@"c:\Users\Barend\Music\Queen\Innuendo");
                NextIfNotStarted();
            }

            //mainDockPanel.Visibility = Visibility.Collapsed;
            //compactPlayer.Visibility = Visibility.Visible;

            //theCompactInfo.labelAlbum.Text = "test";
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            //mainDockPanel.Visibility = Visibility.Visible;
            //compactPlayer.Visibility = Visibility.Collapsed;
        }

        #endregion


    }
}
